﻿/***********************************************************************
 *  filename    : Program.cs
 *  
 *  Description : 本项目用于实现基于Socket异步通讯的Web服务器
 *                主要实现功能:
 *                1. Web-Socket建立，通讯，关闭
 *                2. http协议请求解析，处理，响应报文返回
 *                3. C# Web服务器实现                
 * 
 *  Author       ：zc
 * 
 *  Version      : 0.0.0
 *  
 *  Date         : 2016.1.12
 *  
 *  History      : 2016.1.12
 *                 完成Socket异步通讯处理部分
 *                 完成http服务器静态网页解析引擎， 实现对.html网页文件处理
 *                 添加路径文件解析程序，实现对网页文件的读取
 *                 2016.1.13
 *                 添加对.jpg, .gif, .js, .asp文件的处理
 *                 2016.1.14
 *                 解决频繁刷新时会出现的程序异常结束事件
 *                 1. 监听连接数目过少 16 -> 100
 *                 2. 添加异常处理模块，出现异常时主动关闭当前tcp连接
****************************************************************************/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Text.RegularExpressions;
using System.IO;
using System.Web;


namespace WebServer
{
    class Program
    {
        static void Main(string[] args)
        {
            //创建监听线程
            Thread Listen_thread = new Thread(socket_listen);
            Listen_thread.IsBackground = false;
            Listen_thread.Start();
        }

        private static void socket_listen(object obj)
        {
            Socket server_socket = null;

            //监听的IP地址和端口 作为服务器，绑定的只能是本机Ip地址或者环回地址(不能与系统其它进程端口冲突)
            //如果绑定为本机IP地址，局域网下其它设备可以通过http://host:port来访问当前服务器
            //string host = "127.0.0.1";
            //IPAddress ip = IPAddress.Parse(host);

            int port = 3000;

            //返回当前的所有地址 包括IPv6和IPv4地址 目前只支持IPv4地址 因此索引选择为192.168开头
            IPAddress[] addresslist = Dns.GetHostAddresses(Dns.GetHostName());
            IPAddress ip = (IPAddress)null;
            string default_ip = "192.168";
            for (int i = 0; i < addresslist.GetLength(0); i++)
            {
                ip = (IPAddress)addresslist.GetValue(i);
                if (String.Equals(ip.ToString().Substring(0, default_ip.Length), default_ip))
                    break;
            }

            IPEndPoint ipe = new IPEndPoint(ip, port);

            try
            {
                //新建Socket套接字，绑定在指定的端口并开始监听
                server_socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                server_socket.Bind(ipe);
                server_socket.Listen(100);
                Console.WriteLine("Server Binding at " + ip + ":" + port + "...");
                Console.WriteLine("Wait for connect....");

                while (true)
                {
                    Socket CurrentSocket;

                    //三次握手成功，新建一个连接
                    CurrentSocket = server_socket.Accept();

                    Console.WriteLine("New TCP Socket Create...");

                    //单开一个线程用来处理服务器收发, 并发服务器模式
                    ParameterizedThreadStart tStart = new ParameterizedThreadStart(socket_process);
                    Thread process_thread = new Thread(tStart);
                    process_thread.IsBackground = false;
                    process_thread.Start(CurrentSocket);
                }
            }
            catch(Exception e)
            {
                Console.WriteLine(e);
               // Console.WriteLine("HTTPServer is all Ready work!");
            }
        }

        private static void socket_process(object obj)
        {
            Socket CurrentSocket = (Socket)obj;
            try
            {
                string recvStr = "";
                byte[] recvBytes = new byte[2000];
                int length;
                HTTPServer.HttpProcess hpc = new HTTPServer.HttpProcess();
                hpc.updata_status = false;
                hpc.recv_len = 0;
                hpc.total_len = 0;

                //考虑到Post提交时长度较长时tcp层分片发送，需要等待所有包发送完在处理
                do
                {
                    //获得当前Socket连接传输的数据，并转换为utf-8格式
                    length = CurrentSocket.Receive(recvBytes, recvBytes.Length, 0);
                    recvStr = Encoding.UTF8.GetString(recvBytes, 0, length);

                    Console.WriteLine(recvStr);

                    //http引擎处理，返回获得数据
                    http_engine(recvStr, length, hpc);
                }while(hpc.updata_status);

                //通过socket发送引擎处理后数据
                CurrentSocket.Send(hpc.bs, hpc.bs.Length, 0);
                Console.WriteLine("File Send Finish, Socket Close....\r\n");

                //关闭socket连接
                CurrentSocket.Close();
            }
            catch (Exception exception)
            {
                Console.WriteLine(exception);
                CurrentSocket.Close();
            }
        }

        private static void http_engine(string str, int len, HTTPServer.HttpProcess HttpProcess)
        {
            int pos;

            if (HttpProcess.updata_status)
                dynamic_process(HttpProcess, str);
            else
            {
                //根据\r\n截断，获取http报文首部并转换为小写，方便后续处理
                //Get / HTTP/1.1/r/n
                pos = str.IndexOf("\r\n");
                string str_head = str.Substring(0, pos);
                str_head = str_head.ToLower();

                //根据' '来截断起始行，并赋值给对应参数
                string[] arr = Regex.Split(str_head, @"\s+");
                HttpProcess.hpr = new HTTPServer.HttpPrase();
                HttpProcess.hpr.http_method = arr[0];      // "Get"
                HttpProcess.hpr.url = arr[1];              // "/"
                HttpProcess.hpr.version = arr[2];          // "HTTP/1.1"

                //判断是否有通过ajax要求获得或者提交的动态数据 
                HttpProcess.hpr.dynamic_status = (HttpProcess.hpr.url.IndexOf(".asp") != -1) ? true : false;

                //Console.WriteLine(str);
                if (HttpProcess.hpr.dynamic_status)
                    dynamic_process(HttpProcess, str);
                else
                    static_process(HttpProcess, str);
            }
        }

        //ajax动态数据处理
        private static void dynamic_process(HTTPServer.HttpProcess HttpProcess, string str)
        {
            string send_str = "";
            if (HttpProcess.updata_status == false)
            {
                //解析ajax_methon处理方法
                int PEnd = HttpProcess.hpr.url.IndexOf(".");
                int pStart = HttpProcess.hpr.url.LastIndexOf("/");
                HttpProcess.hpr.dynamic_method = HttpProcess.hpr.url.Substring(pStart + 1, PEnd - pStart - 1);

                if (HttpProcess.hpr.http_method == "post")
                {
                    //解析Content-Length 获得接收数据正文总长度
                    PEnd = str.IndexOf("Content-Length:");
                    str = str.Substring(PEnd);
                    PEnd = str.IndexOf("\r\n");
                    string str_len = str.Substring(16, PEnd - 16);  //获得接收数据总长度！
                    HttpProcess.total_len = int.Parse(str_len);
                    PEnd = str.IndexOf("\r\n\r\n");
                    HttpProcess.updata_status = true;

                    //右移4位，获得正文首字符位置 \r\n\r\n
                    str = str.Substring(PEnd + 4);

                    //对于可能的中文输入，直接用str.length获得的长度与Content-Length不符合
                    //因此取长度用System.Text.Encoding.UTF8.GetBytes(str).Length;中英文获得的长度都正确
                    HttpProcess.recv_len += System.Text.Encoding.UTF8.GetBytes(str).Length;
                }
            }
            else
            {
                HttpProcess.recv_len += System.Text.Encoding.UTF8.GetBytes(str).Length;
            }
           
            if (HttpProcess.recv_len == HttpProcess.total_len)
            {
                HttpProcess.updata_status = false;
            }

            if (HttpProcess.updata_status == false)
            {
                switch (HttpProcess.hpr.dynamic_method)
                {
                    case "aspapply":
                        send_str = create_correct_head(HttpProcess, System.Text.Encoding.UTF8.GetBytes(str).Length);
                        send_str += str;
                        break;
                    case "aspget":
                        send_str = create_correct_head(HttpProcess, HTTPServer.setup.Length);
                        send_str += HTTPServer.setup;
                        break;
                    default:
                        break;
                }
                byte[] send_byte = new byte[send_str.Length];
                send_byte = Encoding.UTF8.GetBytes(send_str);

                Console.WriteLine(send_str);
                HttpProcess.bs = send_byte;
            }
        }

        //静态网页处理
        private static void static_process(HTTPServer.HttpProcess HttpProcess, string str)
        {
            string send_str = "";
            HttpProcess.updata_status = false;
            int pos = str.IndexOf("Authorization:");
            if (HttpProcess.hpr.http_method != "get")
            {
                //错误的请求方式 返回Bad Request
                send_str = HTTPServer.HTTP_BadRequest_Start + HTTPServer.HTTP_ERROR_Head + HTTPServer.HTTP_BadRequest_HTML;
                byte[] send_byte = new byte[send_str.Length];
                send_byte = Encoding.UTF8.GetBytes(send_str);
                HttpProcess.bs = send_byte;
            }
            else if (pos == -1)
            {
                //没有认证消息
                send_str = HTTPServer.http_need_auth_head + HTTPServer.http_need_auth_html;
                byte[] send_byte = new byte[send_str.Length];
                send_byte = Encoding.UTF8.GetBytes(send_str);
                HttpProcess.bs = send_byte;
            }
            else
            {
                do
                {
                    //添加认证部分
                    string auth_str = str.Substring(pos, str.Substring(pos).IndexOf("\r\n")).Substring("Authorization: Basic ".Length);
                    string encode_auth = Encoding.UTF8.GetString(Convert.FromBase64String(auth_str));
                    if (!String.Equals(encode_auth, "admin:admin"))
                    {
                        //错误的请求方式 返回Bad Request
                        send_str = HTTPServer.http_need_auth_head + HTTPServer.http_need_auth_html;
                        byte[] send_byte = new byte[send_str.Length];
                        send_byte = Encoding.UTF8.GetBytes(send_str);
                        HttpProcess.bs = send_byte;
                        break;
                    }

                    //获得当前程序所在的文件夹
                    string url_str = System.AppDomain.CurrentDomain.SetupInformation.ApplicationBase;
                    if (string.Compare(HttpProcess.hpr.url, "/") == 0)
                    {
                        //对于首个请求127.0.0.1:3000/ 返回index.html
                        url_str += "html\\index.html";
                        HttpProcess.hpr.type = "html";
                    }
                    else
                    {
                        //其它请求 如/spring.js 替换为 ...\html\spring.js便于C#查询文件路径
                        url_str = url_str + "html\\" + HttpProcess.hpr.url.Substring(1);
                        url_str = url_str.Replace('/', '\\');
                        int pos_start = url_str.IndexOf('.');

                        //int pos_end = url_str.LastIndexOf('.');
                        //获得当前请求的网页类型 
                        HttpProcess.hpr.type = url_str.Substring(pos_start + 1);
                    }
                    file_process(HttpProcess, send_str, url_str);
                } while(false);
            }
        }

        private static void file_process(HTTPServer.HttpProcess HttpProcess, string bstr, string url)
        {
            if (File.Exists(url))
            {
                try
                {
                    //以文件流的方式打开指定路径内文件
                    using (FileStream fs = new FileStream(url, FileMode.Open, FileAccess.Read))
                    {
                        //StreamReader temp = new StreamReader(fs, Encoding.Default);
                        int fslen = (int)fs.Length;
                        byte[] fbyte = new byte[fslen];
                        int r = fs.Read(fbyte, 0, fslen);
                        fs.Close();

                        bstr = create_correct_head(HttpProcess, fslen);
                        byte[] head_byte = new byte[bstr.Length];
                        head_byte = Encoding.UTF8.GetBytes(bstr);

                        //字符串流合并,生成发送文件 
                        //之前采用的是byte[]->string, string合并， string->byte[]，这种方法读取图片失败
                        //因此修改为，string合并， string->byte[], byte[]合并方式，读取图片成功
                        byte[] send_byte = new byte[head_byte.Length + fbyte.Length];
                        Buffer.BlockCopy(head_byte, 0, send_byte, 0, head_byte.Length);
                        Buffer.BlockCopy(fbyte, 0, send_byte, head_byte.Length * sizeof(byte), fbyte.Length);


                        Console.WriteLine("File Send....");
                        HttpProcess.bs = send_byte;
                    }
                }
                catch
                {
                    Console.WriteLine("文件已打开");
                }
            }
            else
            {
                bstr = HTTPServer.HTTP_NoFound_Start + HTTPServer.HTTP_ERROR_Head + HTTPServer.HTTP_NoFound_HTML;
                byte[] send_byte = new byte[bstr.Length];
                send_byte = Encoding.UTF8.GetBytes(bstr);
                HttpProcess.bs = send_byte;
            }
        }
        private static string create_correct_head(HTTPServer.HttpProcess HttpProcess, long len)
        {
            string head_str;
            string HTTP_Current_Head = HTTPServer.HTTP_OK_Head.Replace("CurrentLength", Convert.ToString(len));

            //if (HttpProcess.hpr.dynamic_status)
            //{
            HTTP_Current_Head = HTTP_Current_Head.Replace("CurrentEncoding", "*");
            //}
            //else
            //{
            //    HTTP_Current_Head = HTTP_Current_Head.Replace("CurrentEncoding", "gzip");
            //}

            //根据不同url需要返回不同的首部类型 具体对比详见http://tool.oschina.net/commons
            switch (HttpProcess.hpr.type)
            {
                case "jpg":
                    HTTP_Current_Head = HTTP_Current_Head.Replace("CurrentType", "application/x-jpg");
                    break;
                case "png":
                    HTTP_Current_Head = HTTP_Current_Head.Replace("CurrentType", "image/png");
                    break;
                case "html":
                    HTTP_Current_Head = HTTP_Current_Head.Replace("CurrentType", "text/html");
                    break;
                case "gif":
                    HTTP_Current_Head = HTTP_Current_Head.Replace("CurrentType", "image/gif");
                    break;
                case "js":
                    HTTP_Current_Head = HTTP_Current_Head.Replace("CurrentType", "application/x-javascript");
                    break;
                case "asp":
                    HTTP_Current_Head = HTTP_Current_Head.Replace("CurrentType", "text/asp");
                    break;
                case "css":
                    HTTP_Current_Head = HTTP_Current_Head.Replace("CurrentType", "text/css");
                    break;
                default:
                    HTTP_Current_Head = HTTP_Current_Head.Replace("CurrentType", "text/html");
                    break;
            }
            head_str = HTTPServer.HTTP_OK_Start + HTTP_Current_Head;
            return head_str;
        }
    }
}
